const std = @import("std");
const debug = std.debug;
const testing = std.testing;

pub const SECONDS_PER_MINUTE = 60;
pub const SECONDS_PER_HOUR = 60 * 60;
pub const SECONDS_PER_DAY = SECONDS_PER_HOUR * 24;

pub const YearTag = enum(u1) {
    normal,
    leap,

    fn new(year: i32) YearTag {
        return if (@rem(year, 4) == 0 and (@rem(year, 100) != 0 or @rem(year, 400) == 0)) .leap else .normal;
    }
};

pub const Year = union(YearTag) {
    normal: i32,
    leap: i32,

    const Self = @This();

    pub fn new(year: i32) Self {
        return switch (YearTag.new(year)) {
            .normal => Self{ .normal = year },
            .leap => Self{ .leap = year },
        };
    }

    pub fn days(self: Self) u16 {
        return switch (self) {
            .normal => 365,
            .leap => 366,
        };
    }

    pub fn seconds(self: Self) i64 {
        return @as(i64, self.days()) * SECONDS_PER_DAY;
    }

    pub fn get(self: Self) i32 {
        return switch (self) {
            .normal => |year| year,
            .leap => |year| year,
        };
    }

    pub fn next(self: Self) Self {
        return Self.new(self.get() + 1);
    }

    pub fn previous(self: Self) Self {
        return Self.new(self.get() - 1);
    }
};

test "new year" {
    try testing.expectEqual(Year.new(2023), Year{ .normal = 2023 });
    try testing.expectEqual(Year.new(2024), Year{ .leap = 2024 });
    debug.print("Passed\n", .{});
}

test "get year" {
    try testing.expectEqual(Year.new(2023).get(), 2023);
    try testing.expectEqual(Year.new(2024).get(), 2024);
    debug.print("Passed\n", .{});
}

test "next year" {
    try testing.expectEqual(Year.new(2023).next(), Year{ .leap = 2024 });
    debug.print("Passed\n", .{});
}

test "last year" {
    try testing.expectEqual(Year.new(2024).previous(), Year{ .normal = 2023 });
    debug.print("Passed\n", .{});
}

pub const Month = enum(u4) {
    january = 1,
    february = 2,
    march = 3,
    april = 4,
    may = 5,
    june = 6,
    july = 7,
    august = 8,
    september = 9,
    october = 10,
    november = 11,
    december = 12,

    const Self = @This();

    pub fn days(self: Self, year: Year) u5 {
        return switch (@enumToInt(self)) {
            1, 3, 5, 7, 8, 10, 12 => 31,
            2 => switch (year) {
                .normal => 28,
                .leap => 29,
            },
            else => 30,
        };
    }

    pub fn seconds(self: Self, year: Year) u32 {
        return @as(u32, self.days(year)) * SECONDS_PER_DAY;
    }

    pub fn next(self: Self) ?Self {
        const num = @enumToInt(self);
        return if (num < 12) @intToEnum(Self, num + 1) else null;
    }

    pub fn previous(self: Self) ?Self {
        const num = @enumToInt(self);
        return if (num > 1) @intToEnum(Self, num - 1) else null;
    }
};

test "get days in month" {
    const year = Year.new(2023);
    const month = Month.february;
    try testing.expectEqual(month.days(year), 28);
    debug.print("Passed\n", .{});
}

test "next month" {
    try testing.expectEqual(Month.june.next(), .july);
    try testing.expectEqual(Month.december.next(), null);
    debug.print("Passed\n", .{});
}

test "last month" {
    try testing.expectEqual(Month.june.previous(), .may);
    try testing.expectEqual(Month.january.previous(), null);
    debug.print("Passed\n", .{});
}

pub const TimeZoneTag = enum(u1) {
    utc,
    offset,
};

pub const Sign = enum(u1) {
    positive,
    negative,
};

const HoursMinutes = struct {
    hours: u4,
    minutes: ?u6,
};

pub const Offset = union(Sign) {
    positive: HoursMinutes,
    negative: HoursMinutes,

    const Self = @This();

    fn new(hours: i8, minutes: ?u6) ?Self {
        if (hours > 12 or hours < -12) {
            return null;
        }
        if (minutes) |m| {
            if (m > 59) return null;
            if (hours == 0 and m == 0) return null;
        } else if (hours == 0) return null;
        if (hours < 0) {
            const h = @intCast(u4, @as(i8, hours) * -1);
            return Self{ .negative = .{ .hours = h, .minutes = minutes } };
        } else {
            return Self{ .positive = .{ .hours = @intCast(u4, hours), .minutes = minutes } };
        }
    }

    fn asSeconds(self: Self) i64 {
        return switch (self) {
            .positive => |ofs| blk: {
                var seconds = @as(i64, ofs.hours) * 3600;
                if (ofs.minutes) |m| seconds += (@as(i64, m) * 60);
                break :blk seconds;
            },
            .negative => |ofs| blk: {
                var seconds = @as(i64, ofs.hours) * 3600;
                if (ofs.minutes) |m| seconds += (@as(i64, m) * 60);
                break :blk seconds * -1;
            },
        };
    }
};

test "new offsets" {
    try testing.expectEqual(Offset.new(-5, null), Offset{ .negative = .{ .hours = 5, .minutes = null } });
    try testing.expectEqual(Offset.new(3, null), Offset{ .positive = .{ .hours = 3, .minutes = null } });
    debug.print("Passed\n", .{});
}

test "as seconds" {
    try testing.expectEqual(Offset.new(-4, 30).?.asSeconds(), -16200);
    try testing.expectEqual(Offset.new(3, null).?.asSeconds(), 10800);
    debug.print("Passed\n", .{});
}

pub const TimeZone = union(TimeZoneTag) {
    utc: void,
    offset: Offset,

    const Self = @This();

    pub fn new(hours: ?i8, minutes: ?u6) ?Self {
        return if (hours) |h| blk: {
            if (h == 0) {
                break :blk .utc;
            } else if (Offset.new(h, minutes)) |ofs| {
                break :blk Self{ .offset = ofs };
            } else {
                break :blk null;
            }
        } else if (minutes) |m| Self{ .offset = Offset.new(0, m).? } else .utc;
    }
};

test "new timezone" {
    const tz = TimeZone.new(-5, null).?;
    try testing.expectEqual(@as(TimeZoneTag, tz), .offset);
    switch (tz) {
        .offset => |ofs| try testing.expectEqual(ofs, Offset{ .negative = .{ .hours = 5, .minutes = null } }),
        else => unreachable,
    }
    debug.print("Passed\n", .{});
}

test "new timezone utc" {
    const tz0 = TimeZone.new(null, null).?;
    const tz1 = TimeZone.new(0, null).?;
    try testing.expectEqual(@as(TimeZoneTag, tz0), .utc);
    try testing.expectEqual(@as(TimeZoneTag, tz1), .utc);
    debug.print("Passed\n", .{});
}

pub const WeekDay = enum(u3) {
    thursday = 0,
    friday,
    saturday,
    sunday,
    monday,
    tuesday,
    wednesday,
};

pub const Comparison = enum {
    gt,
    lt,
    eq,
};

pub const DateTime = struct {
    year: Year,
    month: Month,
    day: u8,
    hour: ?u5,
    minute: ?u6,
    second: ?u6,
    tz: TimeZone,

    const Self = @This();

    pub fn getYear(self: Self) i32 {
        return self.year.get();
    }

    pub fn getOffset(self: Self) ?Offset {
        return switch (self.tz) {
            .utc => null,
            .offset => |ofs| ofs,
        };
    }

    pub fn toTimestamp(self: Self) i64 {
        var seconds: i64 = 0;
        if (self.year.get() < 1970) {
            var year = Year.new(1970);
            while (year.get() != self.year.get()) {
                year = year.previous();
                seconds -= year.seconds();
            }
        } else if (self.year.get() > 1970) {
            var year = Year.new(1970);
            while (year.get() != self.year.get()) {
                seconds += year.seconds();
                year = year.next();
            }
        }
        var month = Month.january;
        while (month != self.month) {
            seconds += month.seconds(self.year);
            month = month.next().?;
        }
        // The days begin numbering with 1, so on the 5th we have had four full
        // days plus some remainder. So we take self.days - 1 for our calculation
        seconds += @as(i64, self.day - 1) * SECONDS_PER_DAY;
        if (self.hour) |h| {
            seconds += @as(i64, h) * 3600;
        }
        if (self.minute) |m| {
            seconds += @as(i64, m) * 60;
        }
        if (self.second) |s| {
            seconds += s;
        }
        if (self.getOffset()) |ofs| seconds -= ofs.asSeconds();
        return seconds;
    }

    pub fn fromTimestamp(ts: i64) Self {
        if (ts < 0) {
            var seconds: i64 = 0;
            var year = Year.new(-1);
            while (seconds > -year.seconds()) {
                seconds -= year.seconds();
                year = year.previous();
            }
            var month: ?Month = Month.december;
            while (month != null) {
                if (month) |m| {
                    if (seconds > m.seconds(year)) break;
                    seconds -= m.seconds(year);
                    month = m.previous();
                }
            }
            var day = month.?.days(year);
            while (day > 0 and seconds < -SECONDS_PER_DAY) {
                seconds -= SECONDS_PER_DAY;
                day -= 1;
            }
            var hours: u5 = 23;
            while (hours >= 0 and seconds < -SECONDS_PER_HOUR) {
                seconds -= SECONDS_PER_HOUR;
                hours -= 1;
            }
            var minutes: u6 = 60;
            while (minutes >= 0 and seconds < -60) {
                seconds -= 60;
                minutes -= 1;
            }
            const second = @intCast(u6, seconds + 60);
            return Self{
                .year = year,
                .month = month.?,
                .day = day,
                .hour = hours,
                .minute = minutes,
                .second = @as(u6, second),
                .tz = .utc,
            };
        } else if (ts > 0) {
            var seconds = ts;
            var year = Year.new(1970);
            while (year.seconds() < seconds) {
                seconds -= year.seconds();
                year = year.next();
            }
            var month = Month.january;
            while (month.seconds(year) < seconds) {
                seconds -= month.seconds(year);
                month = month.next().?;
            }
            const day = @divTrunc(seconds, SECONDS_PER_DAY) + 1;
            seconds = @rem(seconds, SECONDS_PER_DAY);
            const hour = @divTrunc(seconds, SECONDS_PER_HOUR);
            seconds = @rem(seconds, SECONDS_PER_HOUR);
            const minute = @divTrunc(seconds, 60);
            seconds = @rem(seconds, 60);
            return Self{
                .year = year,
                .month = month,
                .day = @intCast(u8, day),
                .hour = @intCast(u5, hour),
                .minute = @intCast(u6, minute),
                .second = @intCast(u6, seconds),
                .tz = .utc,
            };
        } else {
            return Self{
                .year = Year.new(0),
                .month = .january,
                .day = 1,
                .hour = 0,
                .minute = 0,
                .second = 0,
                .tz = .utc,
            };
        }
    }

    pub fn now() Self {
        return Self.fromTimestamp(std.time.timestamp());
    }

    pub fn weekday(self: Self) WeekDay {
        const ts = self.toTimestamp();
        const days = @divTrunc(ts, SECONDS_PER_DAY);
        return @intToEnum(WeekDay, @rem(days, 7));
    }

    pub fn compare(self: Self, other: Self) Comparison {
        const a = self.toTimestamp();
        const b = other.toTimestamp();
        return if (a > b) .gt else if (a < b) .lt else .eq;
    }
};

test "get year" {
    const dt = DateTime{
        .year = Year.new(2023),
        .month = .june,
        .day = 12,
        .hour = 1,
        .minute = 5,
        .second = 14,
        .tz = TimeZone.new(-5, null).?,
    };
    try testing.expectEqual(dt.getYear(), 2023);
    debug.print("Passed\n", .{});
}

test "get offset" {
    const dt = DateTime{
        .year = Year.new(2023),
        .month = .june,
        .day = 12,
        .hour = 1,
        .minute = 5,
        .second = 14,
        .tz = TimeZone.new(-5, null).?,
    };
    try testing.expectEqual(dt.getOffset().?, Offset{ .negative = .{ .hours = 5, .minutes = null } });
    debug.print("Passed\n", .{});
}

test "to timestamp utc" {
    const dt = DateTime{
        .year = Year.new(2023),
        .month = .june,
        .day = 13,
        .hour = 5,
        .minute = 21,
        .second = 22,
        .tz = .utc,
    };
    try testing.expectEqual(dt.toTimestamp(), 1686633682);
    debug.print("Passed\n", .{});
}

test "to timestamp negative offset" {
    const dt = DateTime{
        .year = Year.new(2023),
        .month = .june,
        .day = 13,
        .hour = 0,
        .minute = 21,
        .second = 22,
        .tz = TimeZone.new(-5, null).?,
    };
    try testing.expectEqual(dt.toTimestamp(), 1686633682);
    debug.print("Passed\n", .{});
}

test "conversions" {
    const ts = std.time.timestamp();
    const dt = DateTime.fromTimestamp(ts);
    try testing.expectEqual(dt.toTimestamp(), ts);
    debug.print("Passed\n", .{});
}

test "now" {
    const ts = std.time.timestamp();
    const dt = DateTime.now();
    try testing.expect(ts <= dt.toTimestamp());
    // Make sure they're identical at least to the tens place, since it's possible
    // for them to vary by a second or more depending on system resources
    const a = @divTrunc(ts, 10);
    const b = @divTrunc(dt.toTimestamp(), 10);
    try testing.expectEqual(a, b);
    debug.print("Passed\n", .{});
}

test "get weekday" {
    const dt = DateTime{
        .year = Year.new(2023),
        .month = .june,
        .day = 13,
        .hour = 6,
        .minute = 21,
        .second = 22,
        .tz = .utc,
    };
    try testing.expectEqual(dt.weekday(), .tuesday);
    debug.print("Passed\n", .{});
}

test "get weekday 2" {
    const dt = DateTime{
        .year = Year.new(2023),
        .month = .june,
        .day = 10,
        .hour = 6,
        .minute = 21,
        .second = 22,
        .tz = .utc,
    };
    try testing.expectEqual(dt.weekday(), .saturday);
    debug.print("Passed\n", .{});
}

test "ordering lt" {
    const a = DateTime{
        .year = Year.new(2023),
        .month = .june,
        .day = 10,
        .hour = 6,
        .minute = 21,
        .second = 22,
        .tz = .utc,
    };
    const b = DateTime{
        .year = Year.new(2023),
        .month = .june,
        .day = 10,
        .hour = 6,
        .minute = 21,
        .second = 23,
        .tz = .utc,
    };
    try testing.expectEqual(a.compare(b), .lt);
    debug.print("Passed\n", .{});
}

test "ordering gt" {
    const a = DateTime{
        .year = Year.new(2023),
        .month = .june,
        .day = 10,
        .hour = 6,
        .minute = 21,
        .second = 22,
        .tz = .utc,
    };
    const b = DateTime{
        .year = Year.new(2023),
        .month = .june,
        .day = 10,
        .hour = 6,
        .minute = 21,
        .second = 22,
        .tz = TimeZone.new(1, null).?,
    };
    try testing.expectEqual(a.compare(b), .gt);
    debug.print("Passed\n", .{});
}
